# This file was auto-generated by Fern from our API Definition.

import typing

from .app_variant_output import AppVariantOutput

AddVariantFromBaseAndConfigResponse = typing.Union[AppVariantOutput, typing.Any]
