# This file was auto-generated by Fern from our API Definition.

import enum
import typing

T_Result = typing.TypeVar("T_Result")


class VariantActionEnum(str, enum.Enum):
    """
    An enumeration.
    """

    START = "START"
    STOP = "STOP"

    def visit(
        self, start: typing.Callable[[], T_Result], stop: typing.Callable[[], T_Result]
    ) -> T_Result:
        if self is VariantActionEnum.START:
            return start()
        if self is VariantActionEnum.STOP:
            return stop()
