# This file was auto-generated by Fern from our API Definition.

import enum
import typing

T_Result = typing.TypeVar("T_Result")


class EvaluationType(str, enum.Enum):
    """
    An enumeration.
    """

    HUMAN_A_B_TESTING = "human_a_b_testing"
    SINGLE_MODEL_TEST = "single_model_test"

    def visit(
        self,
        human_a_b_testing: typing.Callable[[], T_Result],
        single_model_test: typing.Callable[[], T_Result],
    ) -> T_Result:
        if self is EvaluationType.HUMAN_A_B_TESTING:
            return human_a_b_testing()
        if self is EvaluationType.SINGLE_MODEL_TEST:
            return single_model_test()
